import numpy as np
from . import options
from . import exceptions

class Domain:
	"""
	Represents some domain of quantification. 

	Attributes:
		- n (int) -- the size of the domain of quantification
	"""
	def __init__(self, size):
		self._n = size

	@property
	def n(self):
		return self._n
	


class DefaultDomain(Domain):
	"""
	A domain whose size is rigidly bound to "options.dom_quant"
	"""
	def __init__(self):
		super(DefaultDomain, self).__init__(options.dom_quant)

	@property
	def n(self):
		return options.dom_quant
	
default_domain = DefaultDomain()
D3 = Domain(3)
D4 = Domain(4)
D5 = Domain(5)

class VarManager:
	"""
	VarManager keeps track of all independent variables in a given system.
	Maps propositional variables and fully saturated predicate variables to indices

	Example: 
	System: unary predicate p and proposition q; domain of individuals = 3
	Values: p(0) p(1) p(2) q
	are mapped to
	Indices: 0    1    2   3
	
	Attributes:
		preds             -- a dictionary mapping predicate indices to the sizes of the domains they depend on
		pred_to_vm_index  -- a dictionary mapping predicate indices to a position (e.g. if a has index 1 and b index 4, a is mapped to position 0 and b to position 1)
		names             -- a dictionary mapping predicate names to their indices 
		memory            -- a list mapping predicate positions to how many bits are required to define this predicate
		offset            -- a list mapping predicate positions to the bit index offset at which they are defined.
	"""
	def __init__(self, preds, names = dict()):
		self.preds = preds 
		self.names = names

		# pred_to_vm_index : dictionary mapping Formula's predicate indices to VarManager's predicate indices
		self.pred_to_vm_index = {pred_idx: vm_idx for vm_idx, pred_idx in enumerate(self.preds.keys())}
		self.linearize()

	def linearize(self):
		"""
		How many independent boolean values to specify the propositional variables.

		propositions: 1 i
		unary predicate: dom_quant
		etc
		"""

		self.memory = [grand_product(deps)  for _, deps in self.preds.items()]
		
		# position in memory of bits devoted to a parcitular predicate
		self.offset = [0]
		for mem in self.memory[:-1]:
			self.offset.append(self.offset[-1] + mem)

	def merge(*vms):
		return VarManager(preds = {k: v for vm in vms for k, v in vm.preds.items()},
						  names = {k: v for vm in vms for k, v in vm.names.items()})

	@property
	def n(self):
		"""
		Number of bits required to specify an assignment 
		"""
		return sum(self.memory)

	def index(self, pred, value_slots):
		if pred in self.preds:
			deps = self.preds[pred]
		else:
			raise exceptions.UnknownPred(pred, self)

		pred_idx = self.pred_to_vm_index[pred]
		offset   = self.offset[pred_idx]

		# to_return  = offset
		# multiplier = 1
		# for slot, dep in zip(value_slots, deps):
		# 	to_return  += slot * multiplier
		# 	multiplier *= dep
		# print(value_slots, deps)
		try:
			return offset + (np.ravel_multi_index(value_slots, deps, order = "F") if deps else 0)
		except ValueError:
			raise Exception("Can't access {} in predicate with domain sizes {}".format(value_slots, deps))


def grand_product(list_numbers):
	"""
	Computes grand product of a list.
	[1, 2, 3] -> 6 (= 2 * 3)
	"""
	to_return = 1
	for nb in list_numbers:
		to_return *= nb
	return to_return
	







import numpy as np
import itertools

from . import options
import exh.utils as utils
from .vars import VarManager
from exh.utils.table import Table
# from formula import Var

class Universe:
	"""	
	Universe generates a set of assignments of truth-values for every predicate (i.e. a world).
	The assigments either cover all the logical space.or just some logical possibilities

	Attributes:
	n      -- number of bits that specify the world (example: propositional varaible a requires 1 bit, unary predicates a(x) as many bits as there are individuaks)
	worlds -- numpy boolean array worlds[i, j] specifies the truth-value of the j-th bit at the i-th world
	vm     -- variable manager ; specifies a mapping from predicates to bit position (example: predicate variable "a" is mapped to "x")
	
	Properties:
	n_worlds -- number of worlds in universe
	"""

	def __init__(self, **kwargs):
		"""
		Keyword arguments:
		f  -- one formula from which to extract the predicates
		vm -- a variable manager object
		fs -- a list of formulas from which to ex
		"""
		
		if "f" in kwargs:
			self.vm = kwargs["f"].vm
		elif "vm" in kwargs:
			self.vm = kwargs["vm"] 
		elif "fs" in kwargs:
			self.vm = VarManager.merge(*[f.vm for f in kwargs["fs"]])

		self.n = self.vm.n

		if "worlds" not in kwargs:
			self.worlds = utils.getAssignment(self.n)
		else:
			self.worlds = kwargs["worlds"]

	@property
	def n_worlds(self):
		return self.worlds.shape[0]
	

	def consistent(self, *fs):

		output = self.evaluate(*fs)

		return np.any(np.min(output, axis = 1))

	# def set(pred, value, **variables):
	# 	if isinstance(pred, Var):
	# 		idx = pred.idx
	# 	else:
	# 		idx = pred

	# 	deps = self.vm.preds[idx]

	# 	# Variables for which no value has been provided
	# 	no_val_vars = list(set(deps.keys()) - set(variables.keys()))
		
	# 	def all_vars_assignment():
	# 		for vals in product(range(options.dom_quant), repeat = len(no_val_vars)):
	# 			d = {var: val for var, val in zip(no_val_vars, vals)}
	# 			d.update(variables)
	# 			yield d


	# 	ko_cols = [self.vm.index(idx, **d) for d in iterator()]

	# 	# We remove all the lines where the values of column does not match value
	# 	reduced_worlds = self.u.worlds[:, ko_cols]
	# 	goal = np.full_like(reduced_worlds, value)

	# 	indices_keep = np.max((goal == reduced_worlds), axis = 1)

	# 	self.worlds = self.worlds[indices_keep, :]


	def entails(self, f1, f2):
		"""Checks if f1 entails f2 in universe"""
		return not self.consistent(f1 & ~f2)

	def equivalent(self, f1, f2):
		output = self.evaluate(f1, f2)
		return np.all(output[:, 0] == output[:, 1])

	def evaluate(self, *fs, **kwargs):
		"""
		Evaluate formules against every world in universe
		"""

		return np.transpose(np.stack([f.evaluate(assignment = self.worlds, vm = self.vm, **kwargs) for f in fs]))


	def name_worlds(self):
		"""
		Returns list of meaningful names to bits, depending on which predicates they set true

		e.g. if Universe is for propositional variable "a", and unary predicate "b" will return
		["a", "b(0)", "b(1)", "b(2)"]
		"""

		def str_tuple(tuple):
			return "({})".format(",".join(list(map(str, t))))

		nvars = self.worlds.shape[1]
		names = [i for i in range(nvars)]
		name_vars = ["A{}".format(key) for key in self.vm.preds.keys()]

		for name, var_idx in self.vm.names.items():
			vm_index = self.vm.pred_to_vm_index[var_idx]
			name_vars[vm_index] = name

		vm_idx_to_deps = list(self.vm.preds.values())

		for i, offset in enumerate(self.vm.offset):
			deps = vm_idx_to_deps[i]
			
			if deps:
				
				for t in np.ndindex(*deps):
					i_col = offset + np.ravel_multi_index(t, deps, order = "F")
					names[i_col] = name_vars[i] + str_tuple(t)

			else:
				names[offset] = name_vars[i]

		return names


	

	

	def restrict(self, indices):
		"""Returns Universe object restricted to the worlds with indices in "indices" argument"""

		return Universe(vm = self.vm, worlds = self.worlds[indices])


	def update(self, var):
		self.vm = VarManager.merge(self.vm, var.vm)
		self.n = self.vm.n
		self.worlds = utils.getAssignment(n)

	def truth_table(self, *fs, **kwargs):
		"""Display a truth-table for formulas fs. Keyword arguments are passed to table (cf exh.utils.table)"""

		output = self.evaluate(*fs)

		table = Table(**kwargs)
		nvars = self.worlds.shape[1]
		nworlds = self.worlds.shape[0]

		# We find the names for the columns
		name_cols = self.name_worlds() + [str(f) for f in fs]
		table.set_header(name_cols)

		# self.worlds: nworlds x nvars
		# output : nworlds x nfs
		combined = np.concatenate([self.worlds, output], axis = 1)

		for row in combined:
			table.add_row(row)

		table.set_strong_col(nvars)
		table.print()	







"""
The classes below find alternatives replacing the root operator by a scalemate ; they preserve the overall structure of the formula
"""

class Scales:
	def __add__(self, other):
		if isinstance(other, ListScales):
			return other + self
		else: 
			return ListScales([self, other])

class SimpleScales(Scales):
	"""
	This class uses the method of "alternative_to" to generate scalemates. It only looks at the type of the prejacent to find its scalemates.
	As such, it cannot do content-based alternative subsitution (e.g. "more than 6" to "more than 7", since both have the same type)
	But it is easy to define since it only demands its scales passed as simple list
	"""

	def __init__(self, scales):
		"""
		Arguments:
		    - scales (list[set[type]]) : a list of scales (seen as sets of types)
		"""
		self.scales = scales


	def alternatives_to(self, prejacents):
		"""
		From a list of prejacents assumed to have the same root operator Op, compute a set of alternatives replacing Op with scalemates

		Arguments:
		    - prejacents (list[Formula]) : a list of formulas
		"""
		if prejacents:
			# Find the scalemates of the prejacent
			example = prejacents[0] # take the first element to find the type of the formulas
			rel_scale = set(type_f for s in self.scales if any(isinstance(example, type_f) for type_f in s) 
			                       for type_f in s if not isinstance(example, type_f))

			# we now need to include scalar replacements
			scalar_alts = []
			for scale_mate in rel_scale:
				for alt_root_fixed in prejacents:
					scalar_alts.append(scale_mate.alternative_to(alt_root_fixed))
			return scalar_alts
		else:
			return []

	def __repr__(self):
		return "SimpleScales({scales})".format(self.scales.__repr__())

class ListScales(Scales):
	"""
	This class takes a list of scales and apply all of them
	"""
	def __init__(self, scales):
		self.scales = scales

	def alternatives_to(self, prejacents):
		return [alt for scale in self.scales for alt in scale.alternatives_to(prejacents)]

	def __add__(self, other):
		if isinstance(other, ListScales):
			return ListScales(self.scales + other.scales)
		else:
			return ListScales(self.scales + [other])


"""
A module for a variety of Generalized Quantifiers and associated scales, including

	- most/more than half
	- modified numerals
"""

import numpy as np
import exh.fol as q



class Most(q.Quantifier):
	plain_symbol = "Most "
	latex_symbol = "\\text{Most }"

	def __init__(self, *args, **kwargs):
		super(Most, self).__init__(*args, **kwargs)
		self.symbol = "Most"
		self.type   = "most"

	def fun(self, results):
		return np.mean(results, axis = 0) > 0.5




class NumeralQuantifier(q.Quantifier):

	def fun(self, results):
		return self.numerosity(np.sum(results, axis = 0))

	def numerosity(self, counts):
		raise Exception("Abstract class NumeralQuantifier can't be instantiated")


class ExactlyQuantifier(NumeralQuantifier):
	plain_symbol = "Exactly "
	latex_symbol = "\\text{{Exactly }}"

	def __init__(self, n, *args, **kwargs):
		self.n = n
		super(ExactlyQuantifier, self).__init__(*args, **kwargs)

		self.plain_symbol += str(self.n)
		self.latex_symbol += str(self.n)

	def numerosity(self, counts):
		return counts == self.n


class MoreThanQuantifier(NumeralQuantifier):
	plain_symbol = "More than "
	latex_symbol = "\\text{{More than }}"

	def __init__(self, n, *args, **kwargs):
		self.n = n
		super(MoreThanQuantifier, self).__init__(*args, **kwargs)

		self.plain_symbol += str(self.n)
		self.latex_symbol += str(self.n)

	def numerosity(self, counts):
		return counts > self.n

class LessThanQuantifier(NumeralQuantifier):
	plain_symbol = "Less than "
	latex_symbol = "\\text{{Less than }}"

	def __init__(self, n, *args, **kwargs):
		self.n = n
		super(LessThanQuantifier, self).__init__(*args, **kwargs)

		self.plain_symbol += str(self.n)
		self.latex_symbol += str(self.n)

	def numerosity(self, counts):
		return counts < self.n


M = q.quantifier_cons(Most)

Mx = M("x")
My = M("y") 
Mz = M("z")

def Exactly(n, var):
	return q.C(lambda formula: ExactlyQuantifier(n, var, formula))

def LessThan(n, var):
	return q.C(lambda formula: LessThanQuantifier(n, var, formula))

def MoreThan(n, var):
	return q.C(lambda formula: MoreThanQuantifier(n, var, formula))


from .quantifier import *

A = quantifier_cons(Universal)
E = quantifier_cons(Existential)


Ax = A("x")
Ex = E("x")

Ay = A("y")
Ey = E("y")

Az = A("z")
Ez = E("z")




def Ax_in_(domain):
	return A("x", domain = domain)

def Ay_in_(domain):
	return A("y", domain = domain)

def Az_in_(domain):
	return A("z", domain = domain)

def Ex_in_(domain):
	return E("x", domain = domain)

def Ey_in_(domain):
	return E("y", domain = domain)

def Ez_in_(domain):
	return E("z", domain = domain)


import numpy as np

# This is a circular import ; better import it at runtime
import exh.alternatives as alternatives
import exh.model        as model
import exh.prop         as prop
import exh.scales       as scale

from exh.utils import jprint
import exh.options as options



class Exhaust:
	"""
	This class orchestrates the computation of exhaustification.
	It computes the alternatives, if they are not provided ; it performs IE and II on them

	Attributes:
		p    (Formula)         -- the prejacent
		alts (list[Formula])   -- the alternatives
		incl (bool)            -- whether IE exhaustification has been computed yet
		incl (bool)            -- whether II exhaustification has been computed yet
		vm   (VariableManager) -- variable manager for prejacent & alternatives
		u    (Universe)        -- universe with all corresponding logical possibilities
	"""
	
	
	def __init__(self, prejacent, alts = None, scales = None, subst = None, extra_alts = []):
		"""
		Arguments
			- prejacent (Formula)       -- the prejacent
			- alts      (list[Formula]) -- stipulated alternatives. If None, alternatives are computed automatically
			- scales                    -- the scales used to compute automatic alternatives
			- subst     (bool)          -- whether subconstituent alternatives should be used
			- extra_alts(list[Formula]) -- if alternatives are computed automatically, add to the already computed alternatives some stipulated ones.
		"""
		# Defining default options dynamically so that users can change options on the fly
		if scales is None:
			scales = options.scales
		elif isinstance(scales, list):
			scales = scale.SimpleScales(scales)
		self.scales = scales

		if subst is None:
			subst = options.sub

		if alts is None: # if no alternative is given, compute them automatically
			self.alts = alternatives.alt(prejacent, scales = scales, subst = subst)
		else:
			self.alts = alts
		self.alts += extra_alts

		self.p = prejacent

		# If there are free variables in the prejacent or the alternatives, they must be saturated with dummy values 
		self.free_vars = set(self.p.free_vars)
		for alt in self.alts:
			self.free_vars.union(set(alt.free_vars))

		self.dummy_vals = {var: 0 for var in self.free_vars}

		self.incl = False
		self.excl = False

		self.vm = model.VarManager.merge(prejacent.vm, *(alt.vm for alt in self.alts))
		self.u  = model.Universe(vm = self.vm)


	def innocently_excludable(self):

		evalSet = [~f for f in self.alts]
		worldsPrejacent = np.squeeze(self.u.evaluate(self.p, no_flattening = True, 
		                                                     variables = self.dummy_vals), # give free variables dummy values
		                             axis = 1) 
		# We restrict ourselves to the worlds where the prejacent is True
		uPrejacent = self.u.restrict(worldsPrejacent)

		if evalSet and uPrejacent.n_worlds != 0:
			self.maximalExclSets         = alternatives.find_maximal_sets(uPrejacent, evalSet, variables = self.dummy_vals)
			self.innocently_excl_indices = np.prod(self.maximalExclSets, axis = 0, dtype = "bool") # innocently_excl_indices[i] is true iff the i-th proposition belongs to every maximal set
		else:
			self.maximalExclSets         = []
			self.innocently_excl_indices = []

		self.excl = True
		return self.innocently_excl_indices

	def innocently_includable(self):

		if not self.excl:
			raise ValueError("Exclusion has not been applied yet.")

		evalNegSet = [~f for f, excludable in zip(self.alts, self.innocently_excl_indices) if excludable] + [self.p]
		evalPosSet = [ f for f, excludable in zip(self.alts, self.innocently_excl_indices) if not excludable]

		worldsStengthenedPrejacent = np.prod(self.u.evaluate(*evalNegSet, no_flattening = True,
		                                                                  variables = self.dummy_vals),
		                                     axis = 1, 
		                                     dtype = "bool")

		# Restricting ourselves to the worlds where the prejacent is true and the negatable alternatives are false.
		uSPrejacent = self.u.restrict(worldsStengthenedPrejacent)
		
		if evalPosSet  and uSPrejacent.n_worlds != 0:
			maximalSets = alternatives.find_maximal_sets(uSPrejacent, evalPosSet, variables = self.dummy_vals)

			# The maximal sets refer to positions in the set of non-excludable alternatives ; we must convert this to position in the whole set of alternatives
			self.maximalInclSets = np.full((len(maximalSets), len(self.alts)), False, dtype = "bool")
			self.maximalInclSets[:, np.logical_not(self.innocently_excl_indices)] = maximalSets

			self.innocently_incl_indices = np.prod(self.maximalInclSets, axis = 0, dtype = "bool")
		else:
			self.maximalInclSets         = []
			self.innocently_incl_indices = []

		self.incl = True
		return self.innocently_incl_indices

	def diagnose(self, display = jprint):
		"""Diplay pertinent information regarding the results of the computation such as maximal sets, IE alternatives, II alternatives"""
		inline_sep = "; "
		list_sep   = "\n   - "

		def sep_fs(fs):
			str_fs = [str(f) for f in fs]
			n_fs   = len(str_fs)

			if n_fs > options.cutoff_inline_to_list:
				return list_sep + list_sep.join(str_fs)
			elif n_fs > 0:
				return inline_sep.join(str_fs)
			else:
				return "nothing"

		if self.excl:
			display("Maximal Sets (excl):")
			for excl in self.maximalExclSets:
				display("{" + inline_sep.join(list(map(str, self.extract_alts(excl)))) + "}") 
			display()
			display("Innocently excludable: ", sep_fs(self.innocently_excl))

		if self.incl:
			display()
			display()
			display("Maximal Sets (incl):")
			for incl in self.maximalInclSets:
				display("{" + inline_sep.join(list(map(str, self.extract_alts(incl)))) + "}") 
			display()
			display("Innocently includable: ", sep_fs(self.innocently_incl))
		display()

	@property
	def innocently_excl(self):
		return self.extract_alts(self.innocently_excl_indices)
	
	@property
	def innocently_incl(self):
		return self.extract_alts(self.innocently_incl_indices)
	

	def extract_alts(self, mask):
		return [alt for alt, included in zip(self.alts, mask) if included]	

class Exh(prop.Operator):
	"""
	This class wraps the class Exhaust into a Formula object, so that it can be evaluated like any Formula object

	Attributes:
		e (Exhaust) -- the object Exhaust that performs the actual computation
	"""

	plain_symbol = "Exh"
	latex_symbol = r"\textbf{Exh}"

	substitutable = False
	
	def __init__(self, child, alts = None, scales = None, subst = None, ii = None, extra_alts = []):
		self.e = Exhaust(child, alts, scales, subst, extra_alts)
		super(Exh, self).__init__(None, child)

		if ii is None:
			self.ii = options.ii_on
		else:
			self.ii = ii
		self.reinitialize()

	def reinitialize(self):
		self.e.p = self.prejacent
		self.ieSet = self.e.innocently_excludable()

		if self.ii:
			self.iiSet = self.e.innocently_includable()
		else:
			self.iiSet = []
		

		self.evalSet = [~f for f, excludable in zip(self.alts, self.ieSet) if excludable] + [f for f, includable in zip(self.alts, self.iiSet) if includable]



	def evaluate_aux(self, assignment, vm, variables = dict(), free_vars = list()):

		evaluanda = [self.children[0]] + self.evalSet
		values = [f.evaluate_aux(assignment, vm, variables) for f in evaluanda]
	
		return np.min(np.stack(values), axis = 0)

	def unpack(self):
		return self.prejacent & prop.And(*self.evalSet)

	@property
	def alts(self):
		return self.e.alts
	

	def vars(self):
		self.vm = model.VarManager.merge(self.children[0].vm, *[alt.vm for alt in self.alts])
		return self.vm

	def diagnose(self, *args, **kwargs):
		self.e.diagnose(*args, **kwargs)

	def __eq__(self, other):
		return isinstance(other, Exh) and (self.children[0] == other.children[0])

	@property
	def prejacent(self):
		return self.children[0]
	

	def copy(self):
		return Exh(self.prejacent, alts = self.alts)

	@classmethod
	def alternative_to(cls, other):
		return cls(other.children[0], alts = other.alts)













# Whether automatic alternatives use subconstituents alternatives by default
sub = True

# Default size of the domain of quantification of quantifiers
dom_quant = 3 

# Whether display is in Latex by default
latex_display = True




from exh.prop    import Or, And
from exh.fol     import Existential, Universal
from exh.scales  import SimpleScales

# Default scalar scales
scales = SimpleScales([{Or, And}, {Existential, Universal}])

# Whether Exh computes innocent inclusion by default
ii_on = False

# Whether automatic alternatives use subconstituents alternatives by default
sub = True

# Whether the prejacent of Exh is an alternative to Exh when "sub" is True ; we don't derive Free Choice if this is set to True
prejacent_alternative_to_exh = False

# The minimmal number of alternatives beyond which the "diagnose" method starts displaying them as bullet point list
cutoff_inline_to_list = 5


import numpy as np
from IPython.core.display import display, HTML


def to_str_list(list_vars):
	return list(map(str, list_vars))

class Table:
	"""
	Helper class to draw tables in HTML and plain text

	Attributes:
		options (dict)
			html (bool)   -- whether to use HTML display or not
			char_col      -- in plain display, character for plain column
			char_bold_col -- in plain display, character for emphasized column
			char_line     -- in plain display, character for line
			row_lines     -- in plain display, whether to insert lines between every row

		header      (list[str])   -- list of cells in header
		rows        (list[list])  -- list of rows ; each row is either a list of strings of the constant HLINE (for horizontal lines)
		strong_col  (list[int])   -- list of indices of vertical lines to display bold (0 means before first column, n_cols means after last_column)
 	"""

	HLINE = 0 # Constant to reprensent line in 
	style ="""
	border: 1px solid black; 
    border-collapse: collapse;
    font-weight: normal; 
	"""

	def __init__(self, **kwargs):
		"""kwargs update class options"""

		self.options = {"html": True, "char_col": "|", "char_bold_col": "#", "char_line": "-", "row_lines": False}
		self.options.update(kwargs)

		self.rows = []
		self.strong_cols = []
		self.header = None




	def set_header(self, list_header):
		""" Set table header to "list_header" argument """
		self.header = to_str_list(list_header)

	def add_row(self, row):
		self.rows.append(to_str_list(row))
		if self.options["row_lines"]:
			self.insert_hline()

	def set_strong_col(self, i):
		# 0 means leftmost border ; ncols + 1 rightmost border
		self.strong_cols.append(i)

	def insert_hline(self):
		self.rows.append(Table.HLINE)




	def print(self, html = None):
		if html is None:
			html = self.options["html"]
		
		if html:
			self.print_html()
		else:
			self.print_plain()

	def print_html(self):

		def cell(i, content):
			style = ""
			if i in self.strong_cols:
				style += "border-left-style: double !important;"
			if i+1 in self.strong_cols:
				style += "border-right-style: double !important;"

			self.add('<th style ="{}">'.format(style))
			self.add(content)
			self.add("</th>")

		self.cache = ""

		self.add('<table style = "{}">'.format(Table.style))

		if self.header is not None:

			self.add("<thead>")
			self.add("<tr>")
			for t in enumerate(self.header):
				cell(*t)
			self.add("</tr>")
			self.add("</thead>")


			for row in self.rows:
				if isinstance(row, list):
					self.add("<tr>")
					for t in enumerate(row):
						cell(*t)
					self.add("</tr>")

		self.add("</table>")

		display(HTML(self.cache))

	def print_plain(self):
		self.cache = ""

		def pad(string, width):
			l = width - len(string) 
			return (l//2 + l % 2) * " " + string + (l//2) * " "

		def print_row(strings, widths):
			for i, (string, width) in enumerate(zip(strings, widths)):
				if i in self.strong_cols:
					self.add(self.options["char_bold_col"])
				else:
					self.add(self.options["char_col"])

				self.add(pad(string, width))

			self.add(self.options["char_col"])
			self.add("\n")

		def print_line(widths):
			self.add((sum(widths) + len(widths) + 1) * self.options["char_line"] + "\n")

		rows_and_headers = ([self.header]  if self.header is not None else []) + [row for row in self.rows if isinstance(row, list)]
		widths = [max(len(row[i]) for row in rows_and_headers) + 2 for i in range(len(rows_and_headers[0]))]

		print_row(self.header, widths)
		print_line(widths)
		
		for row in self.rows:
			if isinstance(row, list):
				print_row(row, widths)

		print(self.cache)
		#self.rows_and_headers

	def add(self, string):
		self.cache += string


__version__ = "1.1"

from .prop    import *
from .model   import *
from .fol     import *
from .exhaust import *
from .scales  import *
# from .worlds import *

# Defining the names that are exported with from exh import *
# __all__ = ["function1", "function2"]



"""
Not implemented yet!
"""

# from formula import Formula


class IteratorType:
	"""
	Iterator over all smallest nodes in the formula that are not of type "type_f"

	Example: in the formula ((a and b) or c) or (d or e), upon input type "or", the iterator returns:
	a and b, c, d, e
	"""
	def iterator_type(self, type_f = None):
		
		if type_f is None:
			type_f = self.__class__

		if isinstance(self, type_f):
			for child in self.children:
				for value in child.iterator_type(type_f):
					yield value
		else:
			yield self










import numpy as np
import itertools

import exh.utils as utils
import exh.model.options as options

### EVALUATION METHODS ###			

class Evaluate:
	
	def evaluate(self, **kwargs):
		"""
		Evaluate the formula with respect to an assignment of values to the propositional and predicate variable

		Arguments:
			vm (VarManager, default = self.vm) -- a variable manager for the variables in the formula
			either
				assignment (np.ndarray[bool]) -- at index i, is the value for the independent variable with index i according to vm
			or
				kwargs (dict) -- for every key, provides the value of the propositional or predicate variable with that name
					if proposition, the value must be boolean
					if n-ary predicate, value must be a boolean numpy array with size (options.dom_quant)^n
			no_flattening (bool, default = False) -- prevent automatic flattening of result if the result is one-dimensional

		Returns:
			np.ndarray[bool] -- Boolean array of shape (n_assignment, dom_quant, ..., dom_quant) specifying for each assignment and values given to free variables
			                                                         <---number of free vars-->
		or 
			bool -- if the latter result is single dimensional
		"""

		if "vm" in kwargs:
			vm = kwargs["vm"]
		else: 
			vm = self.vm

		# If assignment is provided, use it ; otherwise, construct one from keyword arguments
		if "assignment" in kwargs:
			assignment = kwargs["assignment"]
		else:
			assignment = np.full(vm.n, True)

			for var, val in kwargs.items():
				if var in vm.names:
					idx = vm.names[var]
				else:
					continue

				size_domains = vm.preds[idx]
				for t in np.ndindex(*size_domains):
					i = vm.index(idx, t)
					assignment[i] = utils.get(val, t)

			assignment = assignment[np.newaxis, :]

		variables = kwargs.get("variables", dict())
		free_vars = [var for var in kwargs.get("free_vars", self.free_vars) if var not in variables]
		to_return = self.evaluate_aux(assignment, vm, variables = variables, free_vars = free_vars)

		if all(dim == 1 for dim in to_return.shape) and not ("no_flattening" in kwargs and kwargs["no_flattening"]):
			return to_return.item()
		else:
			return to_return


	def evaluate_aux(self, assignment, vm, variables = dict(), free_vars = list()):
		"""
		Auxiliary method for recursion ; evaluates sub-formula (to be overridden by children classes)
		
		Arguments:
			assignment (numpy.ndarray[bool]) -- each line specifies a different assignmen of bit positions to truth values # TODO: rename to worlds
			vm         (VariableManager)     -- variable manager mapping predicate and variable names to bit positions
			variables  (dict[str, int])      -- local assignment of values to variables
			free_vars  (list[str])           -- variables left free in the matrix formula (the formula on which "evaluate" was called)
		
		Returns:
		   np.ndarray[bool] -- Boolean array of shape (n_assignment, dom_quant, ..., dom_quant) specifying for each assignment and values given to free variables
		                                                             <---number of free vars-->
		"""
		raise Exception("evaluate_aux is not been implemented for class {}".format(self.__class__.__name__))







from .formula   import *
from .predicate import *

true  = Truth()
false = Falsity()


a = Pred(0, name = "a")
b = Pred(1, name = "b")
c = Pred(2, name = "c")


if __name__ == "__main__":
	f1 = a & b & ~c
	f2 = a | b & c



import numpy as np
from collections import defaultdict

from exh.model import vars as var
from exh.model import options
import exh.prop as prop

class Quantifier(prop.Formula):
	"""
	Abstract class for quantified formula

	Attributes:
		symbol -- display symbol (obsolete)
		qvar   -- string name of the individual variable of quantification
		domain -- something specifying the number of individuals in the domain
	"""

	substitutable = False
	plain_symbol = "Q"
	latex_symbol = "Q"

	def __init__(self, quant_var, scope, domain = None):
		super(Quantifier, self).__init__(scope)
		self.qvar = quant_var

		try:
			self.free_vars.remove(self.qvar)
		except ValueError:
			pass

		if domain is None:
			domain = var.default_domain
		self.domain = domain

	def display_aux(self, latex):
		return "{symb} {var}, {scope}".format(
			symb = self.latex_symbol if latex else self.plain_symbol,
			var   = self.qvar,
			scope = self.children[0].display_aux(latex)
		) 

	def evaluate_aux(self, assignment, vm, variables = dict(), free_vars = list()):
		return self.fun(np.stack([self.children[0].evaluate_aux(assignment, vm, 
		                                                        dict(variables, **{self.qvar: i}),
		                                                        free_vars) 
		                         for i in range(self.domain.n)],
		                         axis = 0))

	def fun(self, results):
		raise Exception("Evaluation of abstract class Quantifier ; use Universal or Existential class")

	def __eq__(self, other):
		if self.__class__ is other.__class__:
			if self.qvar == other.qvar:
				return self.children[0] == other.children[0]
		return False

	@property
	def scope(self):
		return self.children[0]
	

	def copy(self):
		return self.__class__(self.qvar, self.scope)

	@classmethod
	def alternative_to(cls, other):
		return cls(other.qvar, other.children[0], domain = other.domain)


class Universal(Quantifier):
	plain_symbol = "\u2200"
	latex_symbol = r"\forall"

	def __init__(self, *args, **kwargs):
		super(Universal, self).__init__(*args, **kwargs)

	def fun(self, results):
		return np.min(results, axis = 0)

class Existential(Quantifier):
	plain_symbol = "\u2203"
	latex_symbol = r"\exists"

	def __init__(self, *args, **kwargs):
		super(Existential, self).__init__(*args, **kwargs)

	def fun(self, results):
		return np.max(results, axis = 0)

class C:
	"""
	The following baroque construction allows us to write quantifierd formula in parenthesis-free way:
	Ax > Ey > a | b
	To do this, we twist Python in ways that are not recommendable for other purposes.
	Reasonable usage of the library should not incur any problems.

	The class C is such that "C() > formula37" will return a formula built from "formula37" using function cons
	"""

	def __init__(self, function):
		self.cons = function
		self.bup = function

	def __gt__(self, other): 
		"""
		In python, A > B > C is evaluated as (A > B) and (B > C) ; 
		this is not the semantics we want so we allow A>B to change the value of B before evaluation in conjunct (B>C)
		This is highly unorthodox ; don't try this at home
		"""
		return_val = True
		if isinstance(other, prop.Formula):
			return_val = self.cons(other)
		else:
			f = other.cons # We store a reference to the original function so that the lambda term is not interpreted recursively
			other.cons = lambda formula: self.cons(f(formula))

		self.cons = self.bup # We restore the original function in case chaining has happened
		return return_val

		

def quantifier_cons(constructor):
	""" Returns a C class from a formula constructor """
	def f(var, **kwargs):
		return C(lambda formula: constructor(var, formula, **kwargs))

	return f






"""
This module offers a collection of tools to work with alternatives:
compute alternatives of a formula, find maximal consistent sets, etc.
"""
import numpy as np
from itertools import product
import copy

from . import exhaust
import exh.options  as options
from exh.prop       import Pred, Or, And
from exh.utils      import entails, remove_doubles






############################ MAXIMAL SETS ##############################################


def find_maximal_sets(universe, props, variables = None):
	"""
	Given a set of worlds and a set of propositions, this method returns the maximal sets of propositions that are consistent with one another

	Algorithm:
	focus on the sets S of proposition which are all the propositions true in some world
	return the maximal sets of S

	Arguments:
		universe (Universe)
		props    (list[Formula]) -- set of propositions to compute the maximal sets of

	Returns:
		np.array[bool]           -- returned_value[i, j] is True iff i-th maximal set contains j-th proposition

	"""
	kwargs       = {} if variables is None else {"variables" : variables}
	truth_table  = universe.evaluate(*props, no_flattening = True, **kwargs)
	maximal_sets = []

	# for every world,
	for s in truth_table:

		# test if the set of true proposition in that world is smaller than any of the current maximal sets
		# if yes, go on to the next world
		# if no, remove any smaller set from maximal set and insert
		if any(entails(s, m) for m in maximal_sets):
			continue
		else:
			maximal_sets = [m for m in maximal_sets if not entails(m, s)]
			maximal_sets.append(s)

	return np.stack(maximal_sets) #if maximal_sets else np.full((0, 0), True, dtype = "bool")




############################ ALTERNATIVE GENERATION ##############################################
"""
*alts_aux* performs the automatic generation of alternatives
*alts* adds on top of this simplification of the set of alternatives using braindead heuristics
"""




# @profile
def alt_aux(p, scales, subst):
	"""
	Return alternatives to a formula following a Sauerland-esque algorithm. 
	Specifically, an alternative is anything which can be obtained from the prejacent by sub-constituent replacement ("A" is an alternative to "A or B"), scale replacement ("a or b" is an alternative to "a and b")
	
	Arguments:
		p      (Formula)           -- prejacent
		scales (list[tuple[class]])-- the list of scales in the lexicon
		subst  (bool)              -- whether to take subconstituent alternatives

	Returns:
		list[Formula] -- the alternatives

	"""

	# A predicate has no alternatives
	if isinstance(p, Pred):
		return [p]

	# in case the prejacent is an Exh, its alternative are either stipulated or already computed as "p.alts"
	# So the returned set of alternatives is just {Exh(alt) | alt \in p.alts} (+alt themselved if subst is active)
	if isinstance(p, exhaust.Exh):
		all_alternatives = [p.prejacent]
		all_alternatives.extend(p.alts)

		
		exh_alternatives = [p]
		exh_alternatives.extend(
			[exhaust.Exh(alt, alts = all_alternatives[:i] + all_alternatives[i + 1:]) 
		                    for i, alt in enumerate(all_alternatives) if i != 0 # <--- trick: we don't recompute exhaustification of the prejacent
			]
		) 

		# Somehow, sub-constituent alternative must not be disallowed to derive FC -- something to investigate
		if subst and options.prejacent_alternative_to_exh:
			exh_alternatives.extend(p.alts)

		# exh_alternatives = [exhaust.Exh(alt, alts = all_alternatives[:i] + all_alternatives[i + 1:]) 
		#                     for i, alt in enumerate(all_alternatives)]
		return exh_alternatives

	# GENERAL CASE:



	# Recursively obtain the alternatives of the children nodes of the prejacent 
	children_alternative = [alt_aux(child, scales, subst) for child in p.children]


	root_fixed_alts = [] # alternatives which have the same root operator as the prejacent (ie. a | (b & c) as an alternative to a | (b | c), preserves root or)
	# For every choice of an alternative to a child (C1 x C2 x ... x Cn where Ci are the children's alternatives)
	for t in product(*children_alternative):
		# To avoid problems, the prejacent is copied
		to_append = copy.copy(p)
		to_append.children = t

		# Because exhaust will need to perform computation at initialization, we need to reinitialize. (<= Not sure if this is necessary given that the case Exh is already dealt with)
		to_append.reinitialize()#constructors[p.type](to_append)
		root_fixed_alts.append(to_append)
	# "root_fixed_alts" now contains all alternatives to the current formula that keep the root node the same 
	# e.g. p = a | (b & c) => root_fixed_alts = [a | b, a | c, a | (b | c), a | (b & c)]


	scalar_alts = scales.alternatives_to(root_fixed_alts)
	# # Find the scalemates of the prejacent
	# rel_scale = set(type_f for s in scales if any(isinstance(p, type_f) for type_f in s) 
	#                        for type_f in s if not isinstance(p, type_f))

	# # we now need to include scalar replacements
	# scalar_alts = []
	# for scale_mate in rel_scale:
	# 	for alt_root_fixed in root_fixed_alts:
	# 		scalar_alts.append(scale_mate.alternative_to(alt_root_fixed))

	

	if subst and p.subst: # if sub-constituent are alternatives, add every element of child_alts 
		return root_fixed_alts + scalar_alts + [alt for child_alts in children_alternative for alt in child_alts]
	else:
		return root_fixed_alts + scalar_alts

# @profile
def alt(p, scales = [], subst = False):
	"""
	Simplifies the result of alt_aux for efficiency

	1) Simplify trivial alternatives: A or A -> A, B and B -> B
	2) Remove duplicate alternatives: {A, B, B, A or B} -> {A, B, A or B}
	"""
	return remove_doubles(simplify_alts(alt_aux(p, scales, subst)))
	# return alt_aux(p, scales, subst)




def simplify_alt(alt):
	"""Performs simple heuristics to simplify a formula, namely "A or A" is "A" ; "A and A" is "A" """
	if isinstance(alt, Or) or isinstance(alt, And):
		if len(alt.children) == 2 and alt.children[0] == alt.children[1]:
			return alt.children[0]
	return alt

def simplify_alts(alts):
	"""Applies "simplify_alt" to a list"""
	return list(map(simplify_alt, alts))



	



class UnknownPred(Exception):
	def __init__(self, var, vm):
		self.var = var
		self.vm = vm
		super(UnknownPred, self).__init__("Unknown predicate with index {pred}. Known indices: {vmvars}".format(pred = var,
																									vmvars = list(vm.names.keys())))




import numpy as np
from IPython.display import Math, display, HTML
import itertools

def getAssignment(n):
	"""Returns all possible assignment of values to n independent boolean variables"""
	iterator = [np.array(2**(n-i-1)*(2**(i) * [False] + 2**(i) * [True]), dtype = "bool") for i in range(n)]
	return np.transpose(np.stack(iterator))

def entails(a, b):
	return np.all(np.logical_or(np.logical_not(a), b))

def remove_doubles(fs):
	"""Returns a list of elements from iterable fs, without double values"""

	toReturn = []

	for f in fs:
		if all(f != g for g in toReturn):
			toReturn.append(f)

	return toReturn

def get(array, index_tuple):
	"""Get value from multi-dimensional array "array" at indices specified by tuple "index_tuple" """
	to_return = array
	
	for index in index_tuple:
		to_return = to_return[index]

	return to_return

def add_functions_as_methods(fs):
	def decorator(Class):
		for f in fs:
			setattr(Class, f.__name__, f)
		return Class
	return decorator

def jprint(*args):
	"""Replacement for print in IPython"""
	display(HTML(" ".join(list(map(str, args)))))


def automatic_var_names():
	"""Generator of default variable names"""
	typical_names = ["x{}", "y{}", "z{}"]

	for x in itertools.chain([""], itertools.count()):
		for var in typical_names:
			yield var.format(x)



import exh
from setuptools import setup

long_desc = """\
Exh 
===========================

``Exh`` is a package for computing the exhaustification of logical formulas. 
You can write formulas in a convenient legible syntax. You can compute alternatives automatically. 
You can compute innocently excludable and includable alternatives.   

A introductory tutorial can be found [here](https://github.com/KenyC/Exh/blob/master/examples/tutorial/Tutorial.ipynb).
Other tutorials covering more advanced features and more complicated examples are  available in [the `examples` folder of the GitHub repository](https://github.com/KenyC/Exh/tree/master/examples).
"""

setup(
    name             = "Exh",
    version          = exh.__version__,
    description      = "Computes innocent exclusion/inclusions exhaustivity",
    long_description = long_desc,
    long_description_content_type = "text/markdown",
    url              = "http://github.com/KenyC/Exh",
    author           = "Keny Chatain",
    author_email     = "kchatain@mit.edu",
    license          = "MIT",
    packages         = [
      "exh", 
      "exh.model", 
      "exh.utils", 
      "exh.prop", 
      "exh.exts.gq", 
      "exh.fol"
    ],
    install_requires = [
        "numpy",
        "IPython"
    ],
    zip_safe         = True
)


import exh.model.options as options

### DISPLAY METHODS ####


class Display:
	"""
	Returns string representation of the object, in plain text or Latex
	"""
	def display(self, latex = None):
		if latex is None:
			latex = options.latex_display

		if latex:
			return "${}$".format(self.display_aux(latex))
		else:
			return self.display_aux(latex)

	def display_aux(self, latex):
		raise Exception("Base class Formula cannot be displayed")
		# if latex:
		# 	display_dict = self.__class__.latex_dict
		# else:
		# 	display_dict = self.__class__.plain_dict

		# def paren(typeF, child):
		# 	if (typeF == child.type) or (child.type == "pred") or (child.type == "not"):
		# 		return child.display_aux(display_dict)
		# 	else:
		# 		return "({})".format(child.display_aux(display_dict))

		# if self.type == "not" or self.type == "exh":
		# 	return "{}[{}]".format(display_dict[self.type], self.children[0].display_aux(latex))
		# else:
		# 	return " {type} ".format(type = display_dict[self.type]).join([paren(self.type,	child) for child in self.children])
		
	"""
	Display object, in plain text or LateX
	"""
	def show(self, latex = None):
		if latex is None: # we cannot use "options.latex_display" as default lest we lose dynamic binding
			latex = options.latex_display 

		if latex:
			display(Math(self.display(latex)))
		else:
			print(self.display(latex))




import numpy as np
from collections import defaultdict
import itertools

from IPython.display import Math, display, HTML

import exh.utils         as utils
import exh.model.options as options
import exh.model.vars    as var

from .simplify import IteratorType
from .display  import Display
from .evaluate import Evaluate


class Formula(IteratorType, Display, Evaluate): # Using sub-classing to spread code over multiple files
	"""
	Base class for fomulas

	Class attributes:
		no_parenthesis (bool) -- whether to display the formula with parenthesis around it in conjunctions, coordinations, etc.
		substitutable  (bool) -- Whether sub-formulas of this formulas count as alternatives to it

	Attributes:
		children (list(Formula)) -- sub-formulas
		vm (VariableManager)     -- organizes mapping from predicate and variables name to concrete bit position
	"""

	no_parenthesis = False
	substitutable = True

	def __init__(self, *children):
		self.subst = self.__class__.substitutable
		self.children = children
		self.vars()

		# Free vars are lexically ordered
		self.free_vars = list(set(var for child in self.children for var in child.free_vars)) 
		self.free_vars.sort()

	def reinitialize(self): #only used for Exh, which performs computation at initialization
		pass

	def __and__(self, other):
		return And(self, other)

	def __or__(self, other):
		return Or(self, other)

	def __invert__(self):
		return Not(self)

	def __str__(self):
		return self.display()

	def __repr__(self):
		return self.display()

	def copy(self):
		"""Creates copy of the object (overridden by children's classes"""
		return Formula(*self.children)

	def __eq__(self, other):
		"""Returns true if two formulas are syntactically the same, up to constituent reordering (overridden by children classes)"""
		return self.__class__ is other.__class__


	
	### FORMULA MANIPULATION METHODS ###
	def flatten(self):
		"""
		Turns embedded "or" and "and" in to generalized "or" and "and"
		Ex: a or ((b or c) or d) becomes a or b or c or d 
		"""
		raise Exception("Not implemented yet!")

		if self.type in ["and", "or"]:
			new_children = list(self.iterator_type())
		else:
			new_children = self.children

		return Formula(self.type, *map(lambda c: c.flatten(), new_children))
	

	def simplify(self):
		"""Turns a formula into a quantifier-first, disjunctions of conjunctions formula"""
		raise Exception("Not implemented yet!")

		# Returns all indexes of variables in a conjunctive formula
		def idx_vars(f):
			return [child.idx if "idx" in dir(child) else -1 for child in f.iterator_type("and")]

		if self.type == "or" or self.type == "and":
			simplified_children = [child.simplify() for child in self.iterator_type()]

			if self.type == "or":
				all_children = [grandchild for child in simplified_children for grandchild in child.iterator_type("or")]
				all_children.sort(key = idx_vars)

				return Formula("or", *all_children)

			else:
				all_children = [list(child.iterator_type("or")) for child in simplified_children]
				individual_conjuncts = TODO

				return self
		elif self.type == "neg":
			child = self.children[0]

			if child.type == "or":
				pass
		else:
			return self

	def vars(self):
		"""Returns a VariableManager object for all the variables that occur in the formula"""

		self.vm = var.VarManager.merge(*[c.vars() for c in self.children])
		self.vm.linearize()
		return self.vm

	@classmethod
	def alternative_to(cls, other):
		"""
		Returns an formulat which is an alternative to other with the same children  ; meant to be overriden by subclasses
		Example: Or.alternative_to(a & b) -> a | b
		"""
		raise Exception("alternative_to is not been implemented for class {}".format(cls.__name__))



############### OPERATORS ##############

class Operator(Formula):
	"""
	Base class for associative operators

	Class attributes:
		plain_symbol (str) -- symbol to display in plain text mode (to be overridden by children classes)
		latex_symbol (str) -- symbol to display in LateX mode (to be overridden by children classes)

	Attributes:
		fun (function) -- function to call on subformulas' result to get parent result
	"""


	plain_symbol = "op"
	latex_symbol = "\text{op}"

	def __init__(self, fun, *children):
		super(Operator, self).__init__(*children)
		self.fun = fun
						
	def evaluate_aux(self, assignment, vm, variables = dict(), free_vars = list()):
		"""Stacks subformulas' results and applies fun to it"""
		return self.fun(np.stack([child.evaluate_aux(assignment, vm, variables, free_vars) for child in self.children]))


	def display_aux(self, latex):

		if latex:
			symbol = self.__class__.latex_symbol
		else:
			symbol = self.__class__.plain_symbol

		def paren(child):
			if (self.__class__ is child.__class__) or child.__class__.no_parenthesis:
				return child.display_aux(latex)
			else:
				return "({})".format(child.display_aux(latex))

		if len(self.children) == 1:
			return "{}[{}]".format(symbol, self.children[0].display_aux(latex))
		else:
			return " {type} ".format(type = symbol).join([paren(child) for child in self.children])

	def __eq__(self, other):
		if self.__class__ is other.__class__:
			other_children = list(other.children)
			
			for child1 in self.children:
					
				matches = [i for i, child2 in enumerate(other_children) if child1 == child2]

				if matches:
					other_children.pop(matches[0])
				else:
					return False

			return True
		else:
			return False

	@classmethod
	def alternative_to(cls, other):
		return cls(*other.children)


class And(Operator):
	plain_symbol = "and"
	latex_symbol = r"\land"

	fun_ = lambda array: np.min(array, axis = 0)

	"""docstring for And"""
	def __init__(self, *children):
		super(And, self).__init__(And.fun_, *children)
		

class Or(Operator):
	plain_symbol = "or"
	latex_symbol = r"\lor"
	
	fun_ = lambda array: np.max(array, axis = 0)

	"""docstring for Or"""
	def __init__(self, *children):
		super(Or, self).__init__(Or.fun_, *children)
		
class Not(Operator):
	no_parenthesis = True

	plain_symbol = "not"
	latex_symbol = r"\neg"

	fun_ = lambda x: np.squeeze(np.logical_not(x), axis = 0)

	"""docstring for Not"""
	def __init__(self, child):
		super(Not, self).__init__(Not.fun_, child)




############### TAUTOLOGIES AND ANTILOGIES ########

class Truth(Formula):
	"""docstring for Truth"""
	def __init__(self):
		super(Truth, self).__init__()

	def evaluate_aux(self, assignment, vm, variables = dict(), free_vars = list()):
		return np.ones(assignment.shape[0], dtype = "bool")

	def display_aux(self, latex):
		if latex:
			return r"\textsf{true}"
		else:
			return "true"


class Falsity(Formula):
	"""docstring for Falsity"""
	def __init__(self):
		super(Falsity, self).__init__()

	def evaluate_aux(self, assignment, vm, variables = dict(), free_vars = list()):
		return np.zeros(assignment.shape[0], dtype = "bool")

	def display_aux(self, latex):
		if latex:
			return r"\textsf{true}"
		else:
			return "true"

class Named(Formula):
	def __init__(self, name, child, latex_name = None):
		super(Named, self).__init__(child)
		self.name       = name
		self.latex_name = latex_name if latex_name is not None else self.name

	def evaluate_aux(self, *args, **kwargs):
		return self.children[0].evaluate_aux(*args, **kwargs)

	def display_aux(self, latex):
		if latex:
			return self.latex_name
		else:
			return self.name



import numpy as np

import exh.model.vars     as var
import exh.utils          as utils
import exh.model.options  as options

from .formula import Formula


############### PREDICATE CLASS ################

class Pred(Formula):
	"""
	Class for atomic proposition and predicate variable
	Attributes:
		- name : name for display and evaluation
		- arity : for n-ary predicates, the number of variables that the predicate depends on
		- deps : the name of the default variables that the predicate depends 
		(i.e. when no vars are specified, as in Ax > a, this is what the predicate depends on)
		- idx : an integer that uniquely identifies the predicate
	"""

	no_parenthesis = True
	last_index     = 100 # leaving some offset

	def __init__(self, index = None, name = None, depends = None, domains = None):
		self.name = name

		if index is None:
			index = Pred.last_index
		self.idx  = index

		if self.idx >= Pred.last_index:
			Pred.last_index = self.idx + 1

		if depends is None:
			depends = []
		elif isinstance(depends, str):
			depends = [depends]
		elif isinstance(depends, int):
			depends = [depends]
		self.depends(*depends, domains = domains)


		super(Pred, self).__init__()

		self.free_vars = self.free_vars_

	@property
	def free_vars_(self):
		return sorted(set(self.deps))
	

	def flatten(self):
		return self

	def simplify(self):
		return self

	def display_aux(self, latex):
		if self.deps:
			dep_string = "({})".format(",".join(list(self.deps)))
		else:
			dep_string = ""
		
		if self.name is None:
			return "A{}".format(self.idx) + dep_string
		else:
			return self.name + dep_string

	# @profile
	def evaluate_aux(self, assignment, vm, variables = dict(), free_vars = list()):
		# Not necessary to split by "free_vars", the empty case is just a special case
		# The split avoids generalizing to the worst case.
		if not free_vars: 
			try:
				value_slots = [variables[dep] for dep in self.deps]
			except KeyError as e:
				raise Exception("Predicate {} cannot be evaluated b/c no value for free variable {} was provided".format(self.name, e))

			return assignment[:, vm.index(self.idx, value_slots)]
		else:
			""" 
			P(x, y, z)
			The value of some of these variables are provided by assignment, others are left free
			"""
			raise NotImplementedError
			shape_output           = tuple(options.dom_quant for _ in free_vars)
			vars_not_in_assignment = set(self.deps).difference(set(variables.keys()))

			position_free_vars = np.full(len(free_vars), False)
			for dep in vars_not_in_assignment: 
				position_free_vars[free_vars.index(dep)] = True #Potential for exception if free_vars is misconfigured

			value_slots    = np.full(len(self.deps), 0,     dtype = "int") 
			mask_free_vars = np.full(len(self.deps), False, dtype = "bool") 

			for i, dep in enumerate(self.deps):
				if dep in variables:
					value_slots[i]    = variables[dep]
				else:
					mask_free_vars[i] = True

			output = np.full((len(assignment), *shape_output), True, dtype = "bool")

			for indices in np.ndindex(shape_output):
				value_slots[mask_free_vars] = np.array(indices)[position_free_vars]
				indexing                    = slice(len(assignment)), *indices
				output[indexing]            = assignment[:, vm.index(self.idx, value_slots)]

			return output






	def __call__(self, *variables):
		if len(variables) == self.arity:
			return Pred(self.idx, self.name, variables, domains = self.domains)
		else:
			print("""WARNING: {} variables were provided than the predicate {} depends on ; changing the arity of the predicate to {}. Universe objects will need to be recreated."""
			      .format("More" if len(variables) > self.arity else "Less", self.name, len(variables)))
			self.depends(*variables)
			self.vars() # Recompute variable manager
			self.free_vars = self.free_vars_
			return self

	def __eq__(self, other):
		return super(Pred, self).__eq__(other) and self.idx == other.idx

	def vars(self):
		size_domains = [domain.n for domain in self.domains]

		if self.name is None:
			self.vm = var.VarManager({self.idx: size_domains})
		else:
			self.vm = var.VarManager({self.idx: size_domains}, names = {self.name: self.idx})

		return self.vm




	def depends(self, *depends_on, domains = None):
		if depends_on and isinstance(depends_on[0], int):
			self.arity = depends_on[0]
			self.deps = [var_name for _, var_name in zip(range(self.arity), utils.automatic_var_names())]
		else:
			self.arity = len(depends_on)
			self.deps = depends_on

		if domains is None:
			domains = [var.default_domain for _ in self.deps]
		self.domains = domains
